from urllib.parse import quote
import re
import json
import urllib.request
from .query_parser import QueryParser
from .config import ResourceConfig
from .diagnostics import Diagnostic, DiagnosticTypes

class TsakorpusQueryParser(QueryParser):
    """
    Parses search queries for Tsakorpus-based corpora.
    """

    rxTsakorpusBool = re.compile('[()|,]')

    def build_get_string(self, getParams, config):
        """
        Build a GET string (everything after the ?) from a description
        of the GET parameters in the getParams list.
        """
        termIndexes = self.term_indexes(getParams)
        nWords = len(termIndexes)
        s = 'n_words=' + str(nWords)
        for param in getParams:
            if param[0] in ('word_rel_', 'word_dist_from_', 'word_dist_to_'):
                sfx = '_0'
            else:
                sfx = ''
            s += '&' + param[0] + str(param[1]) + sfx + '=' + quote(str(param[2]))
        for i in termIndexes:
            s += '&lang' + str(i) + '=' + config.search_lang_id
        s += '&page_size=' + str(config.max_hits)
        s += '&precise=on&sort=random&response_format=json&distance_strict=on'
        return s

    def term_query(self, query, config):
        """
        Return list of query parameters for one term or sequence of terms.
        """
        if len(query) >= 2 and query.startswith('"') and query.endswith('"'):
            query = query[1:len(query)-1]
        if len(query) <= 0:
            return Diagnostic(DiagnosticTypes.sru, 10)
        getParams = []
        iTerm = 0
        for term in query.split(' '):
            if len(term) > 0:
                iTerm += 1
                getParams.append(['wf', iTerm, term])
                if iTerm >= 2:
                    # A maximum of one distance constraint per term
                    getParams.append(['word_rel_', iTerm, iTerm-1])
                    getParams.append(['word_dist_from_', iTerm, '1'])
                    getParams.append(['word_dist_to_', iTerm, '1'])
        return getParams

    def binary_bool(self, strOp, operandL, operandR, config):
        if len(operandL) <= 0 or len(operandR) <= 0:
            raise Diagnostic(DiagnosticTypes.sru, 10)
        termsL = self.term_indexes(operandL)
        operandR = self.shift_term_indexes(operandR, max(termsL))
        termsR = self.term_indexes(operandR)
        if strOp == 'AND':
            if len(termsL) > 1 and len(termsR) > 1:
                message = 'Tsakorpus does not support queries that combine several ' \
                          'multi-word sequences with boolean operators.'
                raise Diagnostic(DiagnosticTypes.sru, 48, message=message)
            return operandL + operandR
        elif strOp == 'OR':
            if len(termsL) > 1 or len(termsR) > 1:
                message = 'Tsakorpus does not support queries that combine several ' \
                          'multi-word sequences with boolean operators.'
                raise Diagnostic(DiagnosticTypes.sru, 48, message=message)
            if operandL[0][0] != 'wf' or operandR[0][0] != 'wf':
                raise Diagnostic(DiagnosticTypes.sru, 47)
            if self.rxTsakorpusBool.search(operandL[0][2]) is not None:
                getParamsNew = [('wf', operandL[0][1], '(' + operandL[0][2] + ')|' + operandR[0][2])]
            else:
                getParamsNew = [('wf', operandL[0][1], operandL[0][2] + '|' + operandR[0][2])]
            return getParamsNew
        raise Diagnostic(DiagnosticTypes.sru, 37, details=strOp)

    def not_bool(self, operand, config):
        # TODO: implement
        raise NotImplementedError()

    def send_query(self, strGetParams: str, config: ResourceConfig):
        """
        Send the translated query to the Tsakorpus instance. Return JSON results
        returned by the corpus.
        """
        url = config.resource_base_url.strip('/') + '/search_sent?' + strGetParams
        response = urllib.request.urlopen(url)
        data = response.read()
        encoding = response.info().get_content_charset('utf-8')
        responseJSON = json.loads(data.decode(encoding))
        return responseJSON


if __name__ == '__main__':
    pass

